package net.sergeych.crypto2

import kotlinx.datetime.Instant
import kotlinx.serialization.Serializable
import net.sergeych.bipack.BipackEncoder
import net.sergeych.utools.now

@Serializable
class Seal(
    val publicKey: SigningKey.Public,
    val signature: UByteArray,
    val createdAt: Instant,
    val expiresAt: Instant? = null,
) {

    @Suppress("unused")
    @Serializable
    class SealedData(
        val message: UByteArray,
        val createdAt: Instant?,
        val validUntil: Instant?,
    )

    /**
     * Return true if the seal is correct, see [verify]
     */
    fun isValid(message: UByteArray): Boolean = kotlin.runCatching { verify(message) }.isSuccess

    /**
     * Return Result containing success or error reason if the seal is not correct
     */
    fun check(message: UByteArray): Result<Unit> = kotlin.runCatching { verify(message) }

    /**
     * Check that message is correct for this seal and throws exception if it is not.
     * Note that tampering [createdAt] and [expiresAt] invalidate the seal too.
     *
     * See [check] and [isValid] for non-throwing checks.
     *
     * @throws ExpiredSignatureException
     * @throws IllegalSignatureException
     */
    fun verify(message: UByteArray) {
        val n = now()
        if (createdAt > n) throw IllegalSignatureException("signature's timestamp in the future")
        expiresAt?.let {
            if (n >= it) throw ExpiredSignatureException("signature expired at $it")
        }
        val data = BipackEncoder.encode(SealedData(message, createdAt, expiresAt))
        if (!publicKey.verify(signature, data.toUByteArray()))
            throw IllegalSignatureException()
    }

    companion object {
        operator fun invoke(
            key: SigningKey.Secret, message: UByteArray,
            createdAt: Instant = now(),
            expiresAt: Instant? = null,
        ): Seal {
            val data = BipackEncoder.encode(SealedData(message, createdAt, expiresAt)).toUByteArray()
            return Seal(key.publicKey, key.sign(data), createdAt, expiresAt)
        }
    }
}